#include <stdio.h>
#include <stdlib.h>

#define		POP_SIZE	1
#define		MULTI_EXCH	20
typedef enum { integer, text, method, bool } Type;

typedef struct {
		char name[255];
		Type type;
		char mandatory;
		char deflt[255];
		char examples[255];
		}		Parameter;

char *type_text[4] = { "integer", "string", "specific method", "boolean" };

Parameter list[] =
	{

		{ 
		"max_generations",
 		integer,
		1,
	 	"x",
		"200" 
		},

		{
		"pop_size",
		integer,
		1,
	 	"x",
		"350" 
		},

		{ 
		"random_seed",
 		integer,
		0,
	 	"1",
		"42535" 
		},

		{
		"output.basename",
 		text,
		0,
 		"lilgp",
		"genetic_wonder" 
		},

		{ 
		"output.detail",
 		integer,
		0,
	 	"50",
		"90" 
		},

		{ 
		"output.stt_interval",
	 	integer,
		0,
	 	"1",
		"2" 
		},

		{ 
		"output.bestn",
 		integer,
		0,
	 	"1",
		"10" 
		},

		{ 
		"output.digits",
 		integer,
		0,
	 	"4",
		"6" 
		},

		{ 
		"checkpoint.interval",
	 	integer,
		0,
	 	"1",
		"-1 (= disable)"
		},

		{ 
		"checkpoint.filename",
	 	text,
		0,
 		"gp%06d.ckp",
		"wonder%06d.check" 
		},
		
		{ 
		"checkpoint.compress",
	 	bool,
		0,
		"x",
		"yes" 
		},

		{ 
		"max_nodes",
 		integer, 
		0,
		"x",
		"1024"
		},

		{ 
		"max_depth",
 		integer, 
		0,
		"x",
		"7"
		},

		{ 
		"init.method",
 		method,
		0,
		"half_and_half",
		"grow / full / half_and_half"
		},

		{ 
		"init.depth",
 		text,
		0,
		"2-6",
		"3-10"
		},

		{ 
		"init.random_attempts",
 		integer,
		0,
		"100",
		"333"
		},

		{ 
		"breed_phases",
 		integer,
		1,
		"x",
		"3"
		},

		{ 
		"probabilistic_operators",
 		bool,
		0,
		"no",
		"yes"
		},

		{ 
		"multiple.subpops",
 		integer,
		0,
		"1",
		"12"
		},

		{ 
		"multiple.exch_gen",
 		integer,
		1,
		"x",
		"4"
		},

		{ 
		"multiple.exchanges",
 		integer,
		1,
		"x",
		"12"
		}

	};

int read_input(char *res)
{
int i;
char c;

i = 0;
while ( (c = getchar()) != '\n')
	{
	res[i] = (char) c;
	i++;
	}
res[i] = '\0';

return i;
}


char *user_input (Parameter *p, char *res)
{
int i;

printf ("=====================================================\n");
printf (" Parameter name : %s\n", p->name);
printf (" Type \t\t: %s\n", type_text[p->type]);

if ( p->deflt[0] != 'x' )
	printf (" Default value \t: %s\n", p->deflt);
else
	{
	if (p->mandatory)
		printf ("\t MANDATORY\n");
	else
		printf ("\t optional\n");
	}
printf ("\n");
printf (" Ex. : %s = %s\n", p->name, p->examples);
printf (" Your input : %s = ", p->name);

i = read_input(res);

if (i == 0) 
	{
	if ( p->mandatory )
		{
		printf("\t>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n");
		printf("\t    You failed to give a value for this parameter.\n");
		printf("\t    Please refer to the manual for more information.\n");
		printf("\t>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>\n");
		
		exit (-2);
		}
	else
		{
		/* Using default value */
		strcpy(res, p->deflt);
		}
	}


return res;
}


void exchange_void (FILE *fp, int size, int pops, int count, int bidirectional, char *fromsel, char *tosel)
{
int j;

fprintf(fp, "\n");
fprintf (fp, "#===========================================\n");
fprintf(fp, "#\n");

fprintf(fp, "#\tGeneral exchange\n");

for (j=1; j<=pops; j++)
	{
	fprintf(fp, "exch[%d].from = \n", j);
	fprintf(fp, "exch[%d].fromselect = %s\n", j, fromsel);
	fprintf(fp, "exch[%d].to = \n", j);
	fprintf(fp, "exch[%d].toselect = %s\n", j, tosel);
	fprintf(fp, "exch[%d].count = %d\n", j, count);

	fprintf(fp, "\n");
	}

fprintf(fp, "%s = %d\n", list[MULTI_EXCH].name, j);
fprintf (fp, "#===========================================\n");

}


void exchange_full (FILE *fp, int size, int pops, int count, int bidirectional, char *fromsel, char *tosel)
{
int i, j, k;

fprintf(fp, "\n");
fprintf (fp, "#===========================================\n");
fprintf(fp, "#\n");

fprintf(fp, "#\tFull interconnect %s\n", (bidirectional ? "bidirectional" : "one way") );

printf ("This exchange topology specifies %d individual exchanges.\n", count*pops * (pops - 1));
printf ("So %d individuals will be replaced in each subpopulations.\n", count * (pops - 1));

if ( (count * (pops - 1) ) > size)
	printf (" WARNING : your program will not work! Not enough individuals per subpopulation for theses exchanges!\n");

k = 1;

for (i=1; i<=pops; i++)
	{
	for (j=1; j<i; j++)
		{
		fprintf(fp, "exch[%d].from = %d\n", k, i);
		fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
		fprintf(fp, "exch[%d].to = %d\n", k, j);
		fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
		fprintf(fp, "exch[%d].count = %d\n", k, count);

		fprintf(fp, "\n");

		k++;
		}



	for (j=i+1; j<=pops; j++)
		{
		fprintf(fp, "exch[%d].from = %d\n", k, i);
		fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
		fprintf(fp, "exch[%d].to = %d\n", k, j);
		fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
		fprintf(fp, "exch[%d].count = %d\n", k, count);

		fprintf(fp, "\n");

		k++;
		}

	}
fprintf(fp, "%s = %d\n", list[MULTI_EXCH].name, k-1);
fprintf (fp, "#===========================================\n");

}


void exchange_ring (FILE *fp, int size, int pops, int count, int bidirectional, char *fromsel, char *tosel)
{
int i, k;

fprintf(fp, "\n");
fprintf (fp, "#===========================================\n");
fprintf(fp, "\n");

fprintf(fp, "#\tRing topology %s\n", (bidirectional ? "bidirectional" : "one way") );

printf ("This exchange topology specifies %d individual exchanges.\n", count* (pops - 1) * (bidirectional ? 2 : 1));
printf ("So %d individuals will be replaced in each subpopulations.\n", count  * (bidirectional ? 2 : 1));

if ( (count *  (bidirectional ? 2 : 1)) > size)
	printf (" Incorrect parameters, you program will not work! Not enough individuals per subpopulation!\n");

k = 1;

for (i=1; i<=(pops-1); i++)
	{
	fprintf(fp, "exch[%d].from = %d\n", k, i);
	fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
	fprintf(fp, "exch[%d].to = %d\n", k, i+1);
	fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
	fprintf(fp, "exch[%d].count = %d\n", k, count);

	fprintf(fp, "\n");

	k++;
	}

/* Last one to close the ring */
fprintf(fp, "exch[%d].from = %d\n", k, pops);
fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
fprintf(fp, "exch[%d].to = %d\n", k, 1);
fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
fprintf(fp, "exch[%d].count = %d\n", k, count);

fprintf(fp, "\n");

if (bidirectional)
	{
	for (i=1; i<=(pops-1); i++)
		{
		fprintf(fp, "exch[%d].from = %d\n", k, i+1);
		fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
		fprintf(fp, "exch[%d].to = %d\n", k, i);
		fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
		fprintf(fp, "exch[%d].count = %d\n", k, count);

		fprintf(fp, "\n");

		k++;
		}

	/* Last one to close the ring */
	fprintf(fp, "exch[%d].from = %d\n", k, 1);
	fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
	fprintf(fp, "exch[%d].to = %d\n", k, pops);
	fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
	fprintf(fp, "exch[%d].count = %d\n", k, count);

	fprintf(fp, "\n");
        }

fprintf(fp, "%s = %d\n", list[MULTI_EXCH].name, k-1);
fprintf (fp, "#===========================================\n");

}


void exchange_star (FILE *fp, int size, int pops, int count, int bidirectional, char *fromsel, char *tosel)
{
int i, j, k;

fprintf(fp, "\n");
fprintf (fp, "#===========================================\n");
fprintf(fp, "\n");

fprintf(fp, "#\tStar topology %s\n", (bidirectional ? "bidirectional" : "one way") );

printf ("This exchange topology specifies %d individual exchanges.\n", count * (pops - 1) * (bidirectional ? 2 : 1));

if ( (count * (pops - 1)  * (bidirectional ? 2 : 1)) > size)
	printf (" Incorrect parameters, you program will not work! Not enough individuals per subpopulation!\n");

k = 1;

for (i=2; i<=pops; i++)
	{
	fprintf(fp, "exch[%d].from = %d\n", k, 1);
	fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
	fprintf(fp, "exch[%d].to = %d\n", k, i);
	fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
	fprintf(fp, "exch[%d].count = %d\n", k, count);

	fprintf(fp, "\n");

	k++;
	}


if (bidirectional)
	for (j=2; j<=pops; j++)
		{
		fprintf(fp, "exch[%d].from = %d\n", k, j);
		fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
		fprintf(fp, "exch[%d].to = %d\n", k, 1);
		fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
		fprintf(fp, "exch[%d].count = %d\n", k, count);
        	
		fprintf(fp, "\n");

		k++;
		}

fprintf(fp, "%s = %d\n", list[MULTI_EXCH].name, k-1);
fprintf (fp, "#===========================================\n");

}


void exchange_torus (FILE *fp, int size, int pops, int count, int bidirectional, char *fromsel, char *tosel)
{
int i, j, k, l;
int segments, diameter;
char res[255];

fprintf(fp, "\n");
fprintf (fp, "#===========================================\n");
fprintf(fp, "\n");

fprintf(fp, "#\tTorus topology %s\n", (bidirectional ? "bidirectional" : "one way") );

printf ("Number of segments : ");
fflush(stdout);
read_input(res);
segments = atoi(res);

printf ("Diameter of the segments : ");
fflush(stdout);
read_input(res);
diameter = atoi(res);

if (segments * diameter > pops)
	printf (" Incorrect parameters, you program will fail! You do not have enough subpopulations!\n");


printf ("This exchange topology specifies %d individual exchanges.\n", 2 * count * segments * diameter * (bidirectional ? 2 : 1));
printf ("So %d individuals will be replaced in each subpopulations.\n", count  * 2 * (bidirectional ? 2 : 1));

if ( (count *  (bidirectional ? 2 : 1)) > size)
	printf (" Incorrect parameters, you program will not work! Not enough individuals per subpopulation!\n");

k = 1;
l = 1;

for (j=0; j<segments-1; j++)
	{
	/* Make a segment */
	fprintf(fp, "#\t Segment #%d\n", j);
	fprintf(fp, "#\t=============\n");

	for (i=0; i<diameter-1; i++, l++)
		{
		fprintf(fp, "exch[%d].from = %d\n", k, l);
		fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
		fprintf(fp, "exch[%d].to = %d\n", k, l+1);
		fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
		fprintf(fp, "exch[%d].count = %d\n", k, count);

		fprintf(fp, "\n");

		k++;
		}

	fprintf(fp, "exch[%d].from = %d\n", k, l);
	fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
	fprintf(fp, "exch[%d].to = %d\n", k, l-diameter +1);
	fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
	fprintf(fp, "exch[%d].count = %d\n", k, count);

	fprintf(fp, "\n");

	k++;
	l++;

	/* Connect the segment to the next one */
	fprintf(fp, "#\t Connection to segment #%d\n", j+1);
	fprintf(fp, "#\t=============\n");

	/* Set back the subpopulations count */
	l -= diameter;

	for (i=0; i<diameter; i++, l++)
		{
		fprintf(fp, "exch[%d].from = %d\n", k, l);
		fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
		fprintf(fp, "exch[%d].to = %d\n", k, l+diameter);
		fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
		fprintf(fp, "exch[%d].count = %d\n", k, count);

		fprintf(fp, "\n");

		k++;
		}
	}

/* Last one to close the ring */
fprintf(fp, "#\t Segment #%d\n", j);
fprintf(fp, "#\t=============\n");

for (i=0; i<diameter-1; i++, l++)
	{
	fprintf(fp, "exch[%d].from = %d\n", k, l);
	fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
	fprintf(fp, "exch[%d].to = %d\n", k, l+1);
	fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
	fprintf(fp, "exch[%d].count = %d\n", k, count);

	fprintf(fp, "\n");

	k++;
	}

fprintf(fp, "exch[%d].from = %d\n", k, l);
fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
fprintf(fp, "exch[%d].to = %d\n", k, l-diameter +1);
fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
fprintf(fp, "exch[%d].count = %d\n", k, count);

fprintf(fp, "\n");

k++;
l++;

/* Connect the segment to the next one */
fprintf(fp, "#\t Connection to segment #%d\n", j+1);
fprintf(fp, "#\t=============\n");

/* Set back the subpopulations count */
l -= diameter;

for (i=0; i<diameter; i++, l++)
	{
	fprintf(fp, "exch[%d].from = %d\n", k, l);
	fprintf(fp, "exch[%d].fromselect = %s\n", k, fromsel);
	fprintf(fp, "exch[%d].to = %d\n", k, i+1);
	fprintf(fp, "exch[%d].toselect = %s\n", k, tosel);
	fprintf(fp, "exch[%d].count = %d\n", k, count);

	fprintf(fp, "\n");

	k++;
	}


fprintf(fp, "%s = %d\n", list[MULTI_EXCH].name, k-1);
fprintf (fp, "#===========================================\n");

}


int main (int argc, char **argv)
{
typedef void (*exchange_function)(FILE *, int , int , int, int, char *, char *);
int i, j;
int size;
int count;
int topo;
int stop;
int pops;
int bidirectional;

exchange_function table[7];

char res[255];
char fromsel[255];
char tosel[255];
char filename[255];

FILE *fp;

/* File in the array */
table[0] = exchange_void;
table[1] = exchange_full;
table[2] = exchange_ring;
table[3] = exchange_star;
/* ... */
table[6] = exchange_torus;

if (argc == 2)
	strcpy (filename, argv[1]);
else
	strcpy(filename, "input.file");

printf("Creating a file called %s\n", filename);


fp = fopen (filename, "w");

if (fp == NULL)
	{
	printf ("Unable to open output file (%s)\n", filename);

	exit(-3);
	}


fprintf(fp, "\n");
fprintf(fp, "#\n");
fprintf(fp, "#\n");
fprintf(fp, "# General section\n");
fprintf(fp, "#\n");

for (i=0; i<3; i++)
	{
	user_input(&list[i], res);

	if (i == POP_SIZE)
		size = atoi(res);

	if (res[0] != 'x')
		fprintf(fp, "%s = %s \n", list[i].name, res);
	}

fprintf(fp, "\n");
fprintf(fp, "#\n");
fprintf(fp, "#\n");
fprintf(fp, "# Output parameters\n");
fprintf(fp, "#\n");

for (; i<8; i++)
	{
	user_input(&list[i], res);
	if (res[0] != 'x')
		fprintf(fp, "%s = %s\n", list[i].name, res);
	}

fprintf(fp, "\n");
fprintf(fp, "#\n");
fprintf(fp, "#\n");
fprintf(fp, "# Checkpointing parameters\n");
fprintf(fp, "#\n");

for (; i<11; i++)
	{
	user_input(&list[i], res);
	if (res[0] != 'x')
		fprintf(fp, "%s = %s\n", list[i].name, res);
	}

fprintf(fp, "\n");
fprintf(fp, "#\n");
fprintf(fp, "#\n");
fprintf(fp, "# Size limits\n");
fprintf(fp, "#\n");

for (; i<13; i++)
	{
	user_input(&list[i], res);
	if (res[0] != 'x')
		fprintf(fp, "%s = %s\n", list[i].name, res);
	}

fprintf(fp, "\n");
fprintf(fp, "#\n");
fprintf(fp, "#\n");
fprintf(fp, "# Initialization\n");
fprintf(fp, "#\n");

for (; i<16; i++)
	{
	user_input(&list[i], res);
	if (res[0] != 'x')
		fprintf(fp, "%s = %s\n", list[i].name, res);
	}

fprintf(fp, "\n");
fprintf(fp, "#\n");
fprintf(fp, "#\n");
fprintf(fp, "# Breeding \n");
fprintf(fp, "#\n");

for (; i<18; i++)
	{
	user_input(&list[i], res);

	if ( strcmp (list[i].name, "breed_phases") == 0 )
		stop = atoi(res);

	if (res[0] != 'x')
		fprintf(fp, "%s = %s\n", list[i].name, res);
	}

/* Breeding : additional lines should be created here */
fprintf(fp, "\n");
fprintf (fp, "#===========================================\n");
fprintf(fp, "\n");
for (j=1; j<=stop; j++)
	{
	fprintf(fp, "breed[%d].rate = 0.0\n", j);
	fprintf(fp, "breed[%d].operator = mutation|reproduction|crossover , select=...\n", j);
	fprintf(fp, "\n");
	}

fprintf(fp, "\n");
fprintf (fp, "#===========================================\n");
fprintf(fp, "\n");

fprintf(fp, "\n");
fprintf(fp, "#\n");
fprintf(fp, "#\n");
fprintf(fp, "# Multiple populations \n");
fprintf(fp, "#\n");

for (; i<20; i++)
	{
	user_input(&list[i], res);

	if ( strcmp (list[i].name, "multiple.subpops") == 0 )
		pops = atoi(res);

	if (res[0] != 'x')
		fprintf(fp, "%s = %s\n", list[i].name, res);
	}

if (stop > 1)
	{
	printf (" There are several subpopulations. This program will attempt to generate\n");
	printf (" the exchange statements for you. You can adapt the file afterwards. ;-)\n");
	printf (" Number of individuals for each exchange : ");
	read_input(res);
	count = atoi (res);

	printf (" Which kind of exchange topology would you like to use? (give a number)\n");
	printf ("===================================================================\n");
	printf (" 1) standard form \n");
	printf (" 2) full interconnect\n");
	printf (" 3) ring\n");
	printf (" 4) star\n");
	printf (" 5) cube\t(NYI)\n");
	printf (" 5) hypercube\t(NYI)\n");
	printf (" 6) band\t(NYI)\n");
	printf (" 7) torus\t(NOT bidirectional)\n");
	printf (" Topology number : ");
	read_input(res);
	topo = atoi(res);

	printf (" Do you want the exchanges to be bidirectional [yes/no] : ");
	read_input(res);
	bidirectional = 0;
	if (strcmp(res,"yes") == 0)
		bidirectional = 1;

	printf (" Specify your selection criterion for the source : ");
	read_input(fromsel);

	printf (" Specify your selection criterion for the destination : ");
	read_input(tosel);

	printf ("Generation exchanges ... \n");
	table[topo-1](fp, size, pops, count, bidirectional, fromsel, tosel);
	printf ("Finished generating exchanges.\n");
	}

fprintf(fp, "\n");
fprintf(fp, "#\n");
fprintf(fp, "#\n");
fprintf(fp, "# Application specific options \n");
fprintf(fp, "#\n");


printf ("Closing %s ... ", filename);
fclose(fp);
printf ("done\n");
printf ("\n");
printf ("This program has done all it can do, you have to edit the produced file now:\n");
printf (" - the breeding parameters should still be defined.\n");
printf (" - you can specify you application specific parameters as well (\"app.\" prefix suggested)\n");
printf (" - check whether the exchange parameters are correct.\n");
printf ("\n");
printf ("Good luck\n");
return 1;
}
